from abc import abstractmethod
from dataclasses import dataclass
from typing import Literal


class PromptInterface:

    @abstractmethod
    def generate_question(self, previous_context_body: str) -> str:
        ...


@dataclass
class BasePrompt(PromptInterface):
    body: str

    def generate_question(self, previous_context_body: str) -> str:
        return f'''Question: {self.body}{previous_context_body}'''

    async def async_generate_question(self, previous_context_body: str) -> str:
        return self.generate_question(previous_context_body)


@dataclass
class SpeechPrompt(BasePrompt):
    def generate_question(self, previous_context_body: str) -> str:
        return f'''{self.body}{previous_context_body}'''


@dataclass
class BusinessPrompt(BasePrompt):
    business_context: str

    def generate_question(self, previous_context_body: str) -> str:
        base_question: str = BasePrompt(self.body).generate_question(previous_context_body)
        return f'{base_question}\nBusiness context: {self.business_context}'


@dataclass
class CodePrompt(BusinessPrompt):
    code_context: str
    language: Literal['python']

    def generate_question(self, previous_context_body: str) -> str:
        if self.business_context:
            question: str = BusinessPrompt(self.body, self.business_context).generate_question(previous_context_body)
        else:
            question: str = BasePrompt(self.body).generate_question(previous_context_body)
        return "\n".join(
            (
                question,
                f"\nCode context:\n\n{self.code_context}",
                f"Language: {self.language}"
            )
        )



import json
import os
import wave
import zipfile
from dataclasses import dataclass
from pathlib import Path
from urllib import request

import numpy as np
from pydantic import AnyUrl
from vosk import KaldiRecognizer, SpkModel, Model

from jarvis.config.vosk import VOSK
from jarvis.whisper.inputs import StreamInput, FileInput


@dataclass
class VoskModelDownloader:
    model_url: AnyUrl = f'{VOSK.download_url}/{VOSK.model_name}'
    spk_model_url: AnyUrl = f'{VOSK.download_url}/{VOSK.spk_model_name}'

    # FIXME: Download all
    models_mapping: tuple[tuple[str, AnyUrl]] = (
        (
            VOSK.model_name, f'{VOSK.download_url}/{VOSK.model_name}.zip'
        ),
        (
            VOSK.spk_model_name, f'{VOSK.download_url}/{VOSK.spk_model_name}.zip'
        )
    )

    @classmethod
    def download_models(cls):
        for model_item in cls.models_mapping:

            if not os.path.exists(VOSK.models_path):
                os.mkdir(VOSK.models_path)
            zip_file_path: Path = Path(f'{VOSK.models_path}/{model_item[0]}.zip')
            if not os.path.exists(str(zip_file_path)[:-4]):
                request.urlretrieve(model_item[1], zip_file_path)
                with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
                    zip_ref.extractall(VOSK.models_path)
                os.remove(zip_file_path)


@dataclass
class VoskSpeakerIdentifier:

    @staticmethod
    def __cosine_dist(x: list[float], y: list[float]) -> float:
        nx = np.array(x)
        ny = np.array(y)
        return 1 - np.dot(nx, ny) / np.linalg.norm(nx) / np.linalg.norm(ny)

    @classmethod
    def get_voice_spk(
            cls,
            input_: FileInput | StreamInput
    ) -> list[float]:
        model = Model(lang=VOSK.language)
        spk_model = SpkModel(f'{VOSK.models_path}\\{VOSK.spk_model_name}')
        wf = wave.open(input_.path)
        rec = KaldiRecognizer(model, wf.getframerate())
        rec.SetSpkModel(spk_model)
        while True:
            data = wf.readframes(4000)
            if len(data) == 0:
                break
            rec.AcceptWaveform(data)

        res = json.loads(rec.FinalResult())
        return res['spk']

    @classmethod
    def identify(
            cls,
            input_: FileInput | StreamInput,
            spk_verifier: list[float]
    ) -> bool:
        if cls.__cosine_dist(cls.get_voice_spk(input_), spk_verifier) < VOSK.identify_gate:
            return True
        return False


@dataclass
class VoskSpeaker:
    ...



import logging
from typing import Any, Iterable, Type

from openai import Completion, ChatCompletion
from openai.openai_object import OpenAIObject


from jarvis.config.openai import OPEN_AI
from jarvis.gpt.context import PromptType, ContextHandler
from nltk.tokenize import word_tokenize

logger: logging.Logger = logging.getLogger(__name__)


class _QuestionCompressor:

    @classmethod
    def zip_question(cls, question: str, remove_vowels: bool = False) -> str:
        if remove_vowels:
            question: str = cls.remove_vowels(question)
        tokens: list[str] = word_tokenize(question.lower())
        summary: str = ' '.join(tokens)
        return summary

    @classmethod
    def remove_vowels(cls, text):
        vowels = {'a', 'e', 'i', 'o', 'u', 'A', 'E', 'I', 'O', 'U'}
        words = text.split()
        output_words = []
        for word in words:
            if "-" in word or "'" in word or len(word) < 3:
                output_words.append(word)
                continue
            output_word = word.translate(str.maketrans('', '', ''.join(vowels)))
            output_words.append(output_word)
        output = " ".join(output_words)
        return output


class AbstractCompletion:
    DAVINCI_MODEL: str = 'text-davinci-003'
    CHAT_GPT_MODEL: str = 'gpt-3.5-turbo'

    @classmethod
    def setup_kwargs(
            cls,
            **kwargs
    ) -> dict[str, Any]:
        default_kwargs: dict[str, Any] = {
            'max_tokens': OPEN_AI.answer_tokens_limit,
            'n': 1,
            'stop': None,
            'temperature': 0,
        }
        default_kwargs.update(kwargs)
        return default_kwargs

    @classmethod
    def create(cls, question: str, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    async def acreate(cls, question: str, *args, **kwargs):
        raise NotImplementedError


class DavinciEngine(AbstractCompletion):

    @classmethod
    def create(cls, question: str, *args, **kwargs):
        logger.info(question)
        return Completion.create(
            *args,
            **cls.setup_kwargs(
                engine=OPEN_AI.model_engine,
                prompt=f'{question}\nAnswer language: {OPEN_AI.answer_language}',
                **kwargs
            )
        )

    @classmethod
    async def acreate(cls, question: str, *args, **kwargs):
        logger.info(question)
        return await Completion.acreate(
            *args,
            **cls.setup_kwargs(
                engine=OPEN_AI.model_engine,
                prompt=f'{question}\nAnswer language: {OPEN_AI.answer_language}',
                **kwargs
            )
        )


class ChatGPTEngine(AbstractCompletion):
    messages: list = [{'role': 'system', 'content': f'{OPEN_AI.calibration_message}'}]

    @classmethod
    def create(cls, question: str, *args, **kwargs):
        logger.info(question)
        cls.messages.append({'role': 'user', 'content': question})
        answer = ChatCompletion.create(
            *args,
            **cls.setup_kwargs(
                model=OPEN_AI.model_engine,
                messages=cls.messages,
                **kwargs
            )
        )
        cls.messages.append(answer.choices[0].message)
        return answer

    @classmethod
    async def acreate(cls, question: str, *args, **kwargs):
        logger.info(question)
        cls.messages.append({'role': 'user', 'content': question})
        answer = await ChatCompletion.acreate(
            *args,
            **cls.setup_kwargs(
                model=OPEN_AI.model_engine,
                messages=cls.messages,
                **kwargs
            )
        )
        cls.messages.append(answer.choices[0].message)
        return answer


def engine_type_factory() -> Type[ChatGPTEngine | DavinciEngine]:
    match OPEN_AI.model_engine:
        case AbstractCompletion.CHAT_GPT_MODEL:
            return ChatGPTEngine
        case AbstractCompletion.DAVINCI_MODEL:
            return DavinciEngine
        case _:
            raise NotImplementedError


class SimpleGPT:
    @property
    def engine(self) -> Type[ChatGPTEngine | DavinciEngine]:
        return engine_type_factory()

    @classmethod
    def prepare_answer(cls, answer_list: list[OpenAIObject]) -> str:
        match cls.engine.__name__:
            case ChatGPTEngine.__name__:
                return answer_list[0].message.content
            case DavinciEngine.__name__:
                return ''.join(
                    [x.text for x in answer_list]
                )

    @classmethod
    async def async_process_question(cls, question: str, **kwargs) -> str:
        answer = await cls.engine.acreate(question, **kwargs)
        return cls.prepare_answer(answer.choices)

    @classmethod
    def process_question(cls, question: str, **kwargs) -> str:
        choices = cls.engine.create(question, **kwargs).choices
        return cls.prepare_answer(choices)

    @classmethod
    def process_dialog(
            cls,
            prompts: Iterable[PromptType],
            compress_question: bool = False,
            compress_answer: bool = False,
            q_compressor: Type[_QuestionCompressor] = _QuestionCompressor,
            context_handler: ContextHandler = None,
            **kwargs
    ) -> ContextHandler:

        if context_handler is None:
            context_handler: ContextHandler = ContextHandler()
        for prompt in prompts:
            question: str = context_handler.make_question(prompt)
            try:
                if compress_question:
                    question: str = q_compressor.zip_question(question, kwargs.pop('remove_vowels', False))
                answer: str = cls.process_question(
                    question,
                    **kwargs
                )
                if compress_answer:
                    answer: str = q_compressor.zip_question(answer)
            except ValueError as e:
                match e.args[0]:
                    case 'Question must be provided and not empty.':
                        continue
                    case _:
                        raise e
            else:
                context_handler.register_answer(answer)
        return context_handler

    @classmethod
    async def async_process_dialog(
            cls,
            prompts: Iterable[PromptType],
            compress_question: bool = False,
            compress_answer: bool = False,
            q_compressor: Type[_QuestionCompressor] = _QuestionCompressor,
            context_handler: ContextHandler = None,
            **kwargs
    ) -> ContextHandler:
        if context_handler is None:
            context_handler: ContextHandler = ContextHandler()
        for prompt in prompts:
            question: str = context_handler.make_question(prompt)
            try:
                if compress_question:
                    question: str = q_compressor.zip_question(question, kwargs.pop('remove_vowels', False))
                answer: str = await cls.async_process_question(
                    question,
                    **kwargs
                )
                if compress_answer:
                    answer: str = q_compressor.zip_question(answer)
            except ValueError as e:
                match e.args[0]:
                    case 'Question must be provided and not empty.':
                        continue
                    case _:
                        raise e
            else:
                context_handler.register_answer(answer)
        return context_handler



from typing import Generator
from openai import Audio

from jarvis.config.openai import OPEN_AI
from jarvis.whisper.inputs import FileInput, StreamInput


class WhisperEngine:

    @classmethod
    def transcribe_from_file(cls, input_: FileInput) -> str:
        with input_ as io:
            return Audio.transcribe(OPEN_AI.whisper_model, io).text

    @classmethod
    def translate_from_file(cls, input_: FileInput) -> str:
        with input_ as io:
            return Audio.translate(OPEN_AI.whisper_model, io).text

    @classmethod
    def transcribe_from_stream(cls) -> Generator[str, None, None]:
        for chunk in StreamInput().read_from_stream():
            yield cls.transcribe_from_file(chunk)


    @classmethod
    def translate_from_stream(cls) -> Generator[str, None, None]:
        for chunk in StreamInput().read_from_stream():
            yield cls.translate_from_file(chunk)



from pathlib import Path
from typing import Literal

from pydantic import BaseSettings, AnyUrl


class VoskSettings(BaseSettings):
    """
    https://alphacephei.com/vosk/models
    """
    models_path: Path
    model_name: str = 'vosk-model-small-ru-0.22'
    spk_model_name: str = 'vosk-model-spk-0.4'
    download_url: AnyUrl = 'https://alphacephei.com/vosk/models'
    language: Literal['ru', 'en'] = 'ru'
    identify_gate: float = 0.1

    class Config:
        env_prefix = 'VOSK_'


VOSK: VoskSettings = VoskSettings()



from typing import Literal

import openai
from pydantic import BaseSettings, SecretStr


class OpenAI(BaseSettings):
    secret_key: SecretStr
    model_engine: Literal[
        'text-davinci-003',
        'code-davinci-002',
        'gpt-3.5-turbo',
        'text-ada-001'
    ] = 'gpt-3.5-turbo'
    whisper_model: Literal['whisper-1'] = 'whisper-1'
    max_tokens_limit: int = 4000
    question_tokens_limit: int = 1000
    answer_tokens_limit: int = max_tokens_limit - question_tokens_limit
    answer_language: str = 'ru'
    calibration_message: str = f'You are a helpful sound assistant. Your name - jarvis Answer language: {answer_language}'

    class Config:
        env_prefix = 'OPEN_AI_'


OPEN_AI: OpenAI = OpenAI()

if not openai.api_key:
    openai.api_key = OPEN_AI.secret_key.get_secret_value()



from io import BytesIO
from gtts import gTTS
from pydub import AudioSegment

from jarvis.config.sound import SOUND
from jarvis.gpt.context import ContextHandler


class StreamOutput:

    @classmethod
    def transform(cls, context_handler: ContextHandler) -> AudioSegment:
        fp = BytesIO()
        gTTS(
            text=context_handler.last_answer,
            lang=SOUND.ggts_language,
            tld='com',
            slow=False
        ).write_to_fp(fp)
        fp.seek(0)
        segment = AudioSegment.from_file(fp, 'mp3')
        fp.close()
        return segment


class FileOutput:
    pass



from jarvis.gpt.primitives import BasePrompt, BusinessPrompt, CodePrompt

PromptType = BasePrompt | BusinessPrompt | CodePrompt


class ContextHandler:

    def __init__(self):
        self.current_question: str = ''
        self.last_question: str = ''
        self.last_answer: str = ''
        self.answers: list[str] = []
        self.questions: list[str] = []

    def make_question(self, next_: PromptType) -> str:
        prev_context_body: str = ''
        if self.string_context:
            prev_context_body = f'''\nPrevious context:{self.string_context}'''
        self.current_question = next_.generate_question(prev_context_body)
        self.last_question = f'\nQuestion: {self.current_question}'

        return self.current_question

    def register_answer(self, answer: str) -> None:
        self.questions.append(self.last_question)
        self.last_answer: str = f'\n{answer.lstrip()}'
        self.answers.append(self.last_answer)

    @property
    def string_context(self) -> str:
        return "\n".join(self.questions + self.answers)



import os
from dataclasses import dataclass

import torch

from jarvis.config.silero import SILERO

device = torch.device('cpu')
torch.set_num_threads(4)


@dataclass
class SileroModelDownloader:

    @classmethod
    def download_models(cls):
        if not os.path.isfile(SILERO.model_name):
            torch.hub.download_url_to_file(SILERO.model_download_url, SILERO.model_name)


@dataclass
class SileroSpeakerIdentifier:
    ...


@dataclass
class SileroSpeaker:
    ...


# model = torch.package.PackageImporter(local_file).load_pickle("tts_models", "model")
# model.to(device)

# example_text = 'Привет! Меня зовут Джарвис'
# sample_rate = 48000
# speaker = 'baya'
#
# audio_paths = model.save_wav(text=example_text,
#                              speaker=speaker,
#                              sample_rate=sample_rate)



from pydantic import BaseSettings, AnyUrl


class SileroSettings(BaseSettings):
    """
        https://models.silero.ai/models/tts/ru/v3_1_ru.pt
    """
    download_url: AnyUrl = 'https://models.silero.ai/models/tts'
    language: str = 'ru'
    model_name: str = 'v3_1_ru.pt'

    class Config:
        env_prefix = 'SILERO_'

    @property
    def model_download_url(self):
        return '/'.join((self.download_url, self.language, self.model_name))


SILERO: SileroSettings = SileroSettings()



import speech_recognition as sr

from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import Literal, BinaryIO, Generator
from pydub import AudioSegment

from jarvis.config.sound import SOUND


@dataclass
class FileInput:
    path: Path | str
    format: Literal['mp3', 'wav', 'm4a'] = 'm4a'
    _cur_stream: BinaryIO = None

    def read(self) -> BinaryIO:
        self._cur_stream = open(self.path, 'rb')
        return self._cur_stream

    def __enter__(self):
        return self.read()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._cur_stream.close()


@dataclass
class StreamInput:
    tmp_file_path: Path = Path('./temp_file.wav')
    tmp_file_format: Literal['mp3', 'wav', 'm4a'] = 'wav'
    stop_listen: bool = False
    recognizer = sr.Recognizer()
    recognizer.energy_threshold = SOUND.energy_threshold
    recognizer.pause_threshold = 1

    @staticmethod
    def raw_read(audio_data: sr.AudioData | BytesIO) -> BytesIO:
        if isinstance(audio_data, sr.AudioData):
            return audio_data.get_raw_data()
        return audio_data

    def read(self, audio_data: sr.AudioData | BytesIO) -> FileInput:
        audio_clip = AudioSegment.from_file(BytesIO(audio_data.get_wav_data()), format=self.tmp_file_format)
        audio_clip.export(self.tmp_file_path, format=self.tmp_file_format)
        return FileInput(path=self.tmp_file_path, format=self.tmp_file_format)

    def read_from_stream(self, raw: bool = False) -> Generator[FileInput | BytesIO, None, None]:
        with sr.Microphone(
                device_index=SOUND.device_index,
                sample_rate=SOUND.sample_rate
        ) as source:
            self.recognizer.adjust_for_ambient_noise(source, 1)
            while 1:
                if raw:
                    yield self.raw_read(self.recognizer.listen(source))
                else:
                    yield self.read(self.recognizer.listen(source))
                if self.stop_listen:
                    break

    def record_from_stream(self, duration: int, raw: bool = False) -> FileInput | BytesIO:

        with sr.Microphone(
                device_index=SOUND.device_index,
                sample_rate=SOUND.sample_rate
        ) as source:
            self.recognizer.adjust_for_ambient_noise(source, 1)
            if raw:
                return self.raw_read(self.recognizer.record(source=source, duration=duration))
            else:
                return self.read(self.recognizer.listen(source))



from pydantic import BaseSettings


class SoundSettings(BaseSettings):
    device_index: int | None = None
    energy_threshold: int = 500
    sample_rate: int = 44100
    ggts_language: str = 'ru'
    stop_words: set[str] = {'', }

    class Config:
        env_prefix = 'SOUND_'


SOUND: SoundSettings = SoundSettings()
